/* Copyright (c) 2016 Raffaele Arecchi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package aktrs.rt;

import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Level;

import aktrs.messages.AverageProcessingTimeByClassRequest;
import aktrs.messages.AverageProcessingTimeByClassResponse;
import aktrs.messages.AverageProcessingTimeRequest;
import aktrs.messages.AverageProcessingTimeResponse;
import aktrs.messages.ProfilingMode;
import aktrs.messages.Stop;
import aktrs.messages.ToggleProfilingMessage;
import aktrs.util.Pair;

/**
 * This is the parent class for every actor (local and remote) in aktrs. An
 * instance of this class represents an actor as described in the
 * <a href="https://en.wikipedia.org/wiki/Actor_model">Actor Model (Wikipedia)</a>
 * .<br>
 * <br>
 * Do not instantiate directly this class: instead to create an actor use
 * {@link ActorSystem#summon(String, Class)} <br>
 * This class is expected to be sub-classed to provide custom actors by
 * overriding {@link #onMessage(Object, Actor)} <br>
 * <br>
 * An actor could be local or remote relative to an ActorSystem. The actor is
 * local to an ActorSystem if it is fully managed by this system. The actor is
 * remote to an ActorSystem A if it managed by another ActorSystem B: in this
 * case the real actor lives on B, while a proxy/mirror of the actor for
 * redirecting requests lives on A.
 */
public class Actor implements Runnable {

	private ActorSystem system;
	protected ActorAddress address;
	protected LinkedBlockingQueue<Pair<Actor, Object>> mailbox;
	private Boolean operative = false;
	private Object operativeMutex = new Object();
	private Object initObj;
	private RemoteSystem remoteSystem;

	// profiling attributes
	private Object profilingMutex = new Object();
	private ProfilingMode profiling = ProfilingMode.OFF;
	private Map<Class<?>, BigInteger> processingOverallTime = new HashMap<Class<?>, BigInteger>();
	private Map<Class<?>, BigInteger> processedMessages = new HashMap<Class<?>, BigInteger>();

	public Actor() {
	}

	void setInit(Object obj) {
		this.initObj = obj;
	}

	void setAddress(ActorAddress address) {
		this.address = address;
	}

	void setRemoteSystem(RemoteSystem remoteSystem) {
		this.remoteSystem = remoteSystem;
	}

	void setSystem(ActorSystem actorSystem) {
		this.system = actorSystem;
	}

	/**
	 * @return the ActorSystem in which this actor lives or this actor is a
	 *         proxy/mirror
	 */
	protected ActorSystem getSystem() {
		return system;
	}

	/**
	 * @return the address of this actor
	 */
	public ActorAddress getAddress() {
		return this.address;
	}

	/**
	 * @return true whether this actor is a mirror/proxy for an actor in a
	 *         remote ActorSystem
	 */
	protected boolean isProxyForRemote() {
		return remoteSystem != null;
	}

	/**
	 * @return whether this actor is operative
	 */
	protected boolean isOperative() {
		return operative;
	}

	/**
	 * Override this method if you want to provide custom initialization of this
	 * actor with an initialization object. Initialization is performed before
	 * the actor becomes operative to processing messages into its mailbox.
	 * 
	 * @param initObj
	 *            the initialization object
	 */
	protected void init(Object initObj) {
	}

	/**
	 * Override this method if you want to provide custom size for the actor
	 * mailbox. By default the mailbox size is unbounded. <br>
	 * Please note the if a bounded mailbox is provided, every <b>tell</b>
	 * operation to an actor with a bounded mailbox performed by the
	 * {@link Actor#tell(Actor, Object)} operation is <b>blocking</b> until the
	 * mailbox will be freed to accommodate new incoming messages.
	 * 
	 * @return the actor mailbox size
	 */
	protected int mailBoxSize() {
		return -1;
	}

	/**
	 * Override this method if you want to provide custom behavior after
	 * receiving a {@link Stop} message.
	 * 
	 * @param message
	 *            the stop message
	 */
	protected void manageStop(Stop message) {
	}

	/**
	 * This is the main task this actor does upon receiving a message. Every
	 * class extending the Actor class should override this method to accomplish
	 * the goal it was designed to do.
	 * 
	 * @param message
	 *            the message
	 * @param sender
	 *            the actor who sent the message
	 */
	protected void onMessage(Object message, Actor sender) {
	};

	/**
	 * Override this method to perform custom behavior when an Exception is
	 * thrown processing a message.
	 * 
	 * @param e
	 *            the exception thrown processing a message
	 * @param message
	 *            the message which causes the exception. <b>It could be
	 *            null</b>.
	 * @param sender
	 *            the sender actor
	 */
	protected void onProcessingError(Exception e, Object message, Actor sender) {
	}

	public void run() {
		synchronized (operativeMutex) {
			try {
				if (system.logger.isLoggable(Level.FINE)) {
					system.logger.fine("Starting Actor " + address.toString());
				}
				int mailBoxSize = mailBoxSize();
				if (system.logger.isLoggable(Level.FINE)) {
					system.logger.fine("Actor " + address.toString() + ", mail box size: " + mailBoxSize);
				}
				if (mailBoxSize >= 0) {
					mailbox = new LinkedBlockingQueue<Pair<Actor, Object>>(mailBoxSize);
				} else {
					mailbox = new LinkedBlockingQueue<Pair<Actor, Object>>();
				}
				if (initObj != null) {
					if (system.logger.isLoggable(Level.FINE)) {
						system.logger.fine("Actor " + address.toString() + ", initializing object: " + initObj.toString());
					}
					init(initObj);
				}
				operative = true;
			} catch (Exception e) {
				operative = false;
				system.logger.log(Level.SEVERE, "Error starting Actor " + address.toString(), e);
				return;
			}
			operativeMutex.notifyAll();
		}
		if (system.logger.isLoggable(Level.FINE)) {
			system.logger.fine("Actor " + address.toString() + " is operative");
		}
		while (true) {
			Object message = null;
			Actor sender = null;
			try {
				Pair<Actor, Object> mail = mailbox.take();
				sender = mail.first;
				message = mail.second;
				if (system.logger.isLoggable(Level.FINE)) {
					system.logger.fine("Actor " + address.toString() + ", received message. Sender: " + sender.address.toString() + " , message: "
							+ message.toString());
				}
				if (message instanceof Stop) {
					synchronized (operativeMutex) {
						operative = false;
						mailbox.clear();
						if (system.logger.isLoggable(Level.FINE)) {
							system.logger.fine("Actor " + address.toString() + " is no more operative");
						}
					}
					manageStop((Stop) message);
					return;
				} else if (message instanceof ToggleProfilingMessage) {
					toggleProfiling(((ToggleProfilingMessage) message).mode);
				} else if (message instanceof AverageProcessingTimeByClassRequest) {
					manage((AverageProcessingTimeByClassRequest) message);
				} else if (message instanceof AverageProcessingTimeRequest) {
					manage((AverageProcessingTimeRequest) message);
				}
				long messageProcessingStartTime = new Date().getTime();
				onMessage(message, sender);
				if (profiling == ProfilingMode.ON) {
					long messageProcessingEndTime = new Date().getTime();
					BigInteger newDelta = BigInteger.valueOf(messageProcessingEndTime - messageProcessingStartTime);
					Class<? extends Object> messageClass = message.getClass();
					synchronized (profilingMutex) {
						BigInteger numProcessedMessagesForClass = processedMessages.get(messageClass);
						BigInteger newNumberOfProcessedMessages = numProcessedMessagesForClass != null ? numProcessedMessagesForClass.add(BigInteger.ONE)
								: BigInteger.ONE;
						BigInteger newProcessingOverallTime = numProcessedMessagesForClass != null ? processingOverallTime.get(messageClass).add(newDelta)
								: newDelta;
						processedMessages.put(messageClass, newNumberOfProcessedMessages);
						processingOverallTime.put(messageClass, newProcessingOverallTime);
					}
				}
			} catch (Exception e) {
				if (system.logger.isLoggable(Level.FINE)) {
					system.logger.log(Level.FINE, "Actor " + address.toString() + ", error processing message", e);
				}
				try {
					onProcessingError(e, message, sender);
				} catch (Exception e2) {
					system.logger.log(Level.SEVERE, "Actor " + address.toString() + ", error on handling exception", e2);
				}
			}
		}
	}

	/**
	 * Sends a message to another actor.<br>
	 * Calling this method is <b>blocking</b> until the message is successfully
	 * delivered to the recipient actor mailbox, whether it be local or remote.
	 * This means that processing-order is guaranteed.
	 * 
	 * @param recipientActor
	 *            the recipient actor for the message
	 * @param message
	 *            the message to send
	 */
	public void tell(Actor recipientActor, Object message) {
		if (recipientActor.isProxyForRemote()) {
			if (system.logger.isLoggable(Level.FINE)) {
				system.logger.fine("Actor " + address.toString() + " telling to a remote Actor " + recipientActor.address.toString());
			}
			if (!(message instanceof Serializable)) {
				String errorMsg = "Actor " + address.toString() + " canot tell: message is not serializable: " + message.toString();
				system.logger.severe(errorMsg);
				throw new RuntimeException(errorMsg);
			}
			recipientActor.remoteSystem.trySendMessage(this, recipientActor, (Serializable) message);
		} else {
			if (system.logger.isLoggable(Level.FINE)) {
				system.logger.fine("Actor " + address.toString() + " telling to a local Actor " + recipientActor.address.toString());
			}
			try {
				synchronized (recipientActor.operativeMutex) {
					if (!recipientActor.operative) {
						if (system.logger.isLoggable(Level.FINE)) {
							system.logger.fine("Actor " + recipientActor.address.toString() + " is not operative. Waiting 1 second");
						}
						recipientActor.operativeMutex.wait(1000);
						if (!recipientActor.operative) {
							String errorMsg = "Actor " + address.toString() + " cannot tell: actor " + recipientActor.address.toString() + " is not operative";
							system.logger.severe(errorMsg);
							throw new RuntimeException(errorMsg);
						}
					}
				}
				recipientActor.mailbox.put(new Pair<Actor, Object>(this, message));
			} catch (InterruptedException e) {
				system.logger.log(Level.SEVERE, e.getMessage(), e);
				throw new RuntimeException(e);
			}
		}
	}

	// PROFILING

	private void toggleProfiling(ProfilingMode profiling) {
		if (system.logger.isLoggable(Level.FINE)) {
			system.logger.fine("Actor " + address.toString() + " setting to profiling mode: " + profiling.toString());
		}
		synchronized (profilingMutex) {
			this.profiling = profiling;
		}
	}

	protected void manage(AverageProcessingTimeRequest message) {
		AverageProcessingTimeResponse response = new AverageProcessingTimeResponse();
		response.averageProcessingTime = getAverageProcessingTime();
		response.profiledActorAddress = address;
		tell(system.evoke(message.replyTo), response);
	}

	protected void manage(AverageProcessingTimeByClassRequest message) {
		AverageProcessingTimeByClassResponse response = new AverageProcessingTimeByClassResponse();
		response.averageProcessingTime = getAverageProcessingTime(message.klass);
		response.profiledActorAddress = address;
		response.klass = getClass();
		tell(system.evoke(message.replyTo), response);
	}

	private Double getAverageProcessingTime(Class<?> klass) {
		Double result = new Double(-1);
		synchronized (profilingMutex) {
			BigInteger numOfProcessedMessages = processedMessages.get(klass);
			BigInteger overallTime = processingOverallTime.get(klass);
			if (numOfProcessedMessages != null && overallTime != null) {
				result = new BigDecimal(overallTime).divide(new BigDecimal(numOfProcessedMessages)).doubleValue();
			}

		}
		return result;
	}

	private Double getAverageProcessingTime() {
		Double result = new Double(-1);
		BigInteger numOfProcessedMessages = BigInteger.ZERO;
		BigInteger overallTime = BigInteger.ZERO;
		synchronized (profilingMutex) {
			for (Class<?> klass : processedMessages.keySet()) {
				numOfProcessedMessages = numOfProcessedMessages.add(processedMessages.get(klass));
				overallTime = overallTime.add(processingOverallTime.get(klass));
			}
			if (numOfProcessedMessages.longValue() > 0) {
				result = new BigDecimal(overallTime).divide(new BigDecimal(numOfProcessedMessages)).doubleValue();
			}
		}
		return result;
	}

}
