/* Copyright (c) 2016 Raffaele Arecchi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package aktrs.rt;

import java.io.Serializable;

/**
 * This class represents a universal address for an actor in aktrs. The address
 * is composed by two parts: the first one represents the actor system name in
 * which the actor lives, the second one identifies the actor itself. An address
 * can be represented as a unique string in the format
 * <b>/system-name/actor-name</b>
 */
public class ActorAddress implements Serializable {
	private static final long serialVersionUID = 1L;
	public final String systemId;
	public final String actorId;

	/**
	 * Represent the internal structure of an actor address
	 * 
	 * @param address
	 *            in the format <b>/system-name/actor-name</b>
	 */
	public ActorAddress(String address) {
		super();
		if (!isValid(address)) {
			throw new IllegalArgumentException(address + " is not a valid address");
		}
		String tail = address.substring(1);
		int firstOccurrenceSlash = tail.indexOf('/');
		this.systemId = tail.substring(0, firstOccurrenceSlash);
		this.actorId = tail.substring(firstOccurrenceSlash + 1);
	}

	private boolean isValid(String address) {
		if (address.charAt(0) != '/') {
			return false;
		}
		String tail = address.substring(1);
		int firstOccurrenceSlash = tail.indexOf('/');
		if (firstOccurrenceSlash == 0 || firstOccurrenceSlash == (tail.length() - 1)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "/" + systemId + "/" + actorId;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((actorId == null) ? 0 : actorId.hashCode());
		result = prime * result + ((systemId == null) ? 0 : systemId.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ActorAddress other = (ActorAddress) obj;
		if (actorId == null) {
			if (other.actorId != null)
				return false;
		} else if (!actorId.equals(other.actorId))
			return false;
		if (systemId == null) {
			if (other.systemId != null)
				return false;
		} else if (!systemId.equals(other.systemId))
			return false;
		return true;
	}
}
