/* Copyright (c) 2016 Raffaele Arecchi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package aktrs.rt;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.Socket;
import java.util.logging.Level;

import aktrs.messages.AliveMessage;
import aktrs.messages.DissolveActorMessage;
import aktrs.messages.RemoteMessage;
import aktrs.messages.RemoteResponse;
import aktrs.messages.RemoteResponse.Result;
import aktrs.messages.SummonActorMessage;

public class RemoteSystem {

	private static final int SOCKET_TIMEOUT = 30000;
	// private static final int SOCKET_TIMEOUT = 600000;
	public final String remoteId;
	public final InetAddress address;
	public final int port;
	public final ActorSystem actorSystem;
	private Socket clientSocket;
	private ObjectOutputStream oos;
	private ObjectInputStream ois;
	private Object commandMutex = new Object();

	protected RemoteSystem(ActorSystem actorSystem, String id, InetAddress address, int port) {
		this.actorSystem = actorSystem;
		this.remoteId = id;
		this.address = address;
		this.port = port;
	}

	protected void summon(ActorAddress address, Class<? extends Actor> klass, Serializable initObj) {
		SummonActorMessage summonMessage = new SummonActorMessage();
		summonMessage.address = address;
		summonMessage.klass = klass;
		summonMessage.initInput = initObj;
		sendAndReceiveWithExceptionThrow(summonMessage);
	}

	protected void dissolve(ActorAddress address, String reason) {
		DissolveActorMessage dissolveMessage = new DissolveActorMessage();
		dissolveMessage.address = address;
		dissolveMessage.reason = reason;
		sendAndReceiveWithExceptionThrow(dissolveMessage);
	}

	protected void trySendMessage(Actor sender, Actor recipient, Serializable message) {
		RemoteMessage remoteMessage = new RemoteMessage();
		remoteMessage.message = message;
		remoteMessage.recipientAddress = recipient.address;
		remoteMessage.senderClass = sender.getClass();
		remoteMessage.senderAddress = sender.address;
		remoteMessage.remoteReplyPort = actorSystem.getListeningPort();
		sendAndReceiveWithExceptionThrow(remoteMessage);
	}

	private void sendAndReceiveWithExceptionThrow(Serializable message) {
		if (actorSystem.logger.isLoggable(Level.FINE)) {
			actorSystem.logger.fine("Sending message " + message.toString() + " to Remote System: " + toString());
		}
		synchronized (commandMutex) {
			try {
				if (clientSocket == null || !isConnected(clientSocket)) {
					if (actorSystem.logger.isLoggable(Level.FINE)) {
						actorSystem.logger.fine("Remote System: " + toString() + " is not connected");
					}
					reconnect();
				}
				oos.writeObject(message);
				oos.flush();
				RemoteResponse readObject = (RemoteResponse) ois.readObject();
				if (actorSystem.logger.isLoggable(Level.FINE)) {
					actorSystem.logger.fine("Received response: " + readObject.toString() + " from Remote System: " + toString());
				}
				if (readObject.result.equals(RemoteResponse.Result.ERROR)) {
					actorSystem.logger.severe(readObject.description);
					throw new RuntimeException(readObject.description);
				}
			} catch (Exception e) {
				disconnect();
				actorSystem.logger.log(Level.SEVERE, e.getMessage(), e);
				throw new RuntimeException(e);
			}
		}
	}

	private boolean isConnected(Socket clientSocket2) {
		try {
			oos.writeObject(AliveMessage.create());
			oos.flush();
			RemoteResponse readObject = (RemoteResponse) ois.readObject();
			return readObject.result == Result.OK;
		} catch (Exception e) {
			return false;
		}
	}

	private void reconnect() throws IOException {
		actorSystem.logger.info("Reconnecting to Remote System: " + toString());
		disconnect();
		clientSocket = new Socket(address, port);
		clientSocket.setSoTimeout(SOCKET_TIMEOUT);
		oos = new ObjectOutputStream(clientSocket.getOutputStream());
		ois = new ObjectInputStream(clientSocket.getInputStream());
		actorSystem.logger.info("Reconnected to Remote System: " + toString());
	}

	protected void disconnect() {
		actorSystem.logger.info("Disconnecting Remote System: " + toString());
		try {
			if (clientSocket != null) {
				clientSocket.close();
				clientSocket = null;
			}
			if (oos != null) {
				oos.close();
				oos = null;
			}
			if (ois != null) {
				ois.close();
				ois = null;
			}
		} catch (IOException e) {
		}
		actorSystem.logger.info("Disconnected from Remote System: " + toString());
	}

	@Override
	public String toString() {
		return remoteId + "[" + address.toString() + ":" + port + "]";
	}

}
